@@ -16,6 +16,7 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 from libqtile.layout.base import Layout, _ClientList
+from libqtile.log_utils import logger
 
 
 class _Column(_ClientList):
@@ -37,7 +38,8 @@ class _Column(_ClientList):
             dict(
                 heights=[self.heights[c] for c in self.clients],
                 split=self.split,
-            ))
+            )
+        )
         return info
 
     def toggle_split(self):
@@ -67,11 +69,14 @@ class _Column(_ClientList):
 
     def __str__(self):
         cur = self.current
-        return "_Column: " + ", ".join([
-            "[%s: %d]" % (c.name, self.heights[c])
-            if c == cur else "%s: %d" % (c.name, self.heights[c])
-            for c in self.clients
-        ])
+        return "_Column: " + ", ".join(
+            [
+                "[%s: %d]" % (c.name, self.heights[c])
+                if c == cur
+                else "%s: %d" % (c.name, self.heights[c])
+                for c in self.clients
+            ]
+        )
 
 
 class Columns(Layout):
@@ -104,34 +109,46 @@ class Columns(Layout):
         Key([mod, "control"], "k", lazy.layout.grow_up()),
         Key([mod, "control"], "h", lazy.layout.grow_left()),
         Key([mod, "control"], "l", lazy.layout.grow_right()),
+        Key([mod, "shift", "control"], "h", lazy.layout.swap_column_left()),
+        Key([mod, "shift", "control"], "l", lazy.layout.swap_column_right()),
         Key([mod], "Return", lazy.layout.toggle_split()),
         Key([mod], "n", lazy.layout.normalize()),
     """
+
     defaults = [
-        ("name", "columns", "Name of this layout."),
-        ("border_focus", "#881111", "Border colour for the focused window."),
-        ("border_normal", "#220000", "Border colour for un-focused windows."),
-        ("border_focus_stack", "#881111",
-         "Border colour for the focused window in stacked columns."),
-        ("border_normal_stack", "#220000",
-         "Border colour for un-focused windows in stacked columns."),
+        ("border_focus", "#881111", "Border colour(s) for the focused window."),
+        ("border_normal", "#220000", "Border colour(s) for un-focused windows."),
+        (
+            "border_focus_stack",
+            "#881111",
+            "Border colour(s) for the focused window in stacked columns.",
+        ),
+        (
+            "border_normal_stack",
+            "#220000",
+            "Border colour(s) for un-focused windows in stacked columns.",
+        ),
         ("border_width", 2, "Border width."),
         ("border_on_single", False, "Draw a border when there is one only window."),
         ("margin", 0, "Margin of the layout (int or list of ints [N E S W])."),
-        ("margin_on_single", -1, "Margin when only one window. `-1` means use `margin`."),
+        (
+            "margin_on_single",
+            None,
+            "Margin when only one window. (int or list of ints [N E S W])",
+        ),
         ("split", True, "New columns presentation mode."),
         ("num_columns", 2, "Preferred number of columns."),
         ("grow_amount", 10, "Amount by which to grow a window/column."),
         ("fair", False, "Add new windows to the column with least windows."),
-        ("insert_position", 0,
-         "Position relative to the current window where new ones are inserted "
-         "(0 means right above the current window, 1 means right after)."),
-        ("wrap_focus_columns", True,
-         "Wrap the screen when moving focus across columns."),
-        ("wrap_focus_rows", True,
-         "Wrap the screen when moving focus across rows."),
-        ("wrap_focus_stacks", True,
-         "Wrap the screen when moving focus across stacked."),
+        (
+            "insert_position",
+            0,
+            "Position relative to the current window where new ones are inserted "
+            "(0 means right above the current window, 1 means right after).",
+        ),
+        ("wrap_focus_columns", True, "Wrap the screen when moving focus across columns."),
+        ("wrap_focus_rows", True, "Wrap the screen when moving focus across rows."),
+        ("wrap_focus_stacks", True, "Wrap the screen when moving focus across stacked."),
     ]
 
     def __init__(self, **config):
@@ -145,13 +162,19 @@ class Columns(Layout):
         c.columns = [_Column(self.split, self.insert_position)]
         return c
 
+    def get_windows(self):
+        clients = []
+        for c in self.columns:
+            clients.extend(c.clients)
+        return clients
+
     def info(self):
         d = Layout.info(self)
         d["clients"] = []
         d["columns"] = []
         for c in self.columns:
             cinfo = c.info()
-            d["clients"].extend(cinfo['clients'])
+            d["clients"].extend(cinfo["clients"])
             d["columns"].append(cinfo)
         d["current"] = self.current
         return d
@@ -177,6 +200,9 @@ class Columns(Layout):
         return c
 
     def remove_column(self, col):
+        if len(self.columns) == 1:
+            logger.warning("Trying to remove all columns.")
+            return
         idx = self.columns.index(col)
         del self.columns[idx]
         if idx <= self.current:
@@ -230,10 +256,9 @@ class Columns(Layout):
         if len(self.columns) == 1 and (len(col) == 1 or not col.split):
             if not self.border_on_single:
                 border = 0
-            if self.margin_on_single > -1:
+            if self.margin_on_single is not None:
                 margin_size = self.margin_on_single
-        width = int(
-            0.5 + col.width * screen_rect.width * 0.01 / len(self.columns))
+        width = int(0.5 + col.width * screen_rect.width * 0.01 / len(self.columns))
         x = screen_rect.x + int(0.5 + pos * screen_rect.width * 0.01 / len(self.columns))
         if col.split:
             pos = 0
@@ -241,17 +266,11 @@ class Columns(Layout):
                 if client == c:
                     break
                 pos += col.heights[c]
-            height = int(
-                0.5 + col.heights[client] * screen_rect.height * 0.01 / len(col))
+            height = int(0.5 + col.heights[client] * screen_rect.height * 0.01 / len(col))
             y = screen_rect.y + int(0.5 + pos * screen_rect.height * 0.01 / len(col))
             client.place(
-                x,
-                y,
-                width - 2 * border,
-                height - 2 * border,
-                border,
-                color,
-                margin=margin_size)
+                x, y, width - 2 * border, height - 2 * border, border, color, margin=margin_size
+            )
             client.unhide()
         elif client == col.cw:
             client.place(
@@ -261,7 +280,8 @@ class Columns(Layout):
                 screen_rect.height - 2 * border,
                 border,
                 color,
-                margin=margin_size)
+                margin=margin_size,
+            )
             client.unhide()
         else:
             client.hide()
@@ -278,8 +298,9 @@ class Columns(Layout):
 
     def focus_next(self, win):
         """Returns the next client after 'win' in layout,
-           or None if there is no such client"""
-        # First: try to get next window in column of win
+        or None if there is no such client"""
+        # First: try to get next window in column of win (self.columns is non-empty)
+        # pylint: disable=undefined-loop-variable
         for idx, col in enumerate(self.columns):
             if win in col:
                 nxt = col.focus_next(win)
@@ -293,8 +314,9 @@ class Columns(Layout):
 
     def focus_previous(self, win):
         """Returns the client previous to 'win' in layout.
-           or None if there is no such client"""
-        # First: try to focus previous client in column
+        or None if there is no such client"""
+        # First: try to focus previous client in column (self.columns is non-empty)
+        # pylint: disable=undefined-loop-variable
         for idx, col in enumerate(self.columns):
             if win in col:
                 prev = col.focus_previous(win)
@@ -316,7 +338,7 @@ class Columns(Layout):
                 self.current = (self.current - 1) % len(self.columns)
         else:
             if self.current > 0:
-                self.current = (self.current - 1)
+                self.current = self.current - 1
         self.group.focus(self.cc.cw, True)
 
     def cmd_right(self):
@@ -325,7 +347,7 @@ class Columns(Layout):
                 self.current = (self.current + 1) % len(self.columns)
         else:
             if len(self.columns) - 1 > self.current:
-                self.current = (self.current + 1)
+                self.current = self.current + 1
         self.group.focus(self.cc.cw, True)
 
     def want_wrap(self, col):
@@ -479,3 +501,18 @@ class Columns(Layout):
                 col.heights[client] = 100
             col.width = 100
         self.group.layout_all()
+
+    def swap_column(self, src, dst):
+        self.columns[src], self.columns[dst] = self.columns[dst], self.columns[src]
+        self.current = dst
+        self.group.layout_all()
+
+    def cmd_swap_column_left(self):
+        src = self.current
+        dst = src - 1 if src > 0 else len(self.columns) - 1
+        self.swap_column(src, dst)
+
+    def cmd_swap_column_right(self):
+        src = self.current
+        dst = src + 1 if src < len(self.columns) - 1 else 0
+        self.swap_column(src, dst)
